// install necessary modules
var express = require("express");
var bodyParser = require("body-parser");
var app = express();
var player = require("play-sound")(opts={});

app.use(bodyParser.urlencoded({ extended: false } ));
app.use(express.static('static'));

// listen to localhost:3000
var server = app.listen(3000, function () {
	console.log('listening on port 3000');
    matching = Object.assign({}, matchingDefault);
    for(var k in matching){
      if(matching.hasOwnProperty(k)) {
        matching[k] = matchingDefault[k];
      }
    }
});

var io = require("socket.io")(server);
var Myo = require("myo");

//connect with MYO armband
Myo.connect();
Myo.on("connected", function(data, timestamp) {
    console.log("Myo successfully connected. Data: " + JSON.stringify(data) + ". Timestamp: " + timestamp + ".");
	Myo.setLockingPolicy("none");
    //console.log(Myo);
});

// useful variables
var punchTime = 0;  
var onTestPunch = false;
var onTestRoll = "none";
var onTestPitch = "none";
var onTestYaw = "none";
var firstOrientationAngles;
var orientationRedo = true;
var isCalibrating = false;
var isLogging = false;
var isControlling = false;

// array storing the path of sound file
var act = {
    "forward":"./sound/forward.wav",
    "forwardleft":"./sound/forwardleft.wav",
    "forwardright":"./sound/forwardright.wav",
    "backward":"./sound/backward.wav",
    "backwardleft":"./sound/backwardleft.wav",
    "backwardright":"./sound/backwardright.wav",
    "bowup":"./sound/bowup.wav",
    "bowdown":"./sound/bowdown.wav",
    "punchleft":"./sound/punchleft.wav",
    "punchright":"./sound/punchright.wav",
    "open":"./sound/open.wav",
    "close":"./sound/close.wav",
    "rotateleft":"./sound/rotateleft.wav",
    "rotateright":"./sound/rotateright.wav",
    "rotatestop":"./sound/rotatestop.wav",
    "turnleft":"./sound/turnleft.wav",
    "turnright":"./sound/turnright.wav",
    "reset":"./sound/reset.wav",
    "play": "playarray"
};

// array storing default gesture control setting
var matchingDefault = {
    "double_tap":act.play,
    "fingers_spread":act.rotatestop,
    "fist":act.backward,
    "fistup":act.bowup,
    "fistdown":act.bowdown,
    "fistright":act.backwardright,
    "fistleft":act.backwardleft,
    "fistclockwise":act.open,
    "fistanticlockwise":act.close,
    "fistpunch":"null",
    "wave_in":act.forward,
    "wave_inup":"null",
    "wave_indown":"null",
    "wave_inright":act.forwardright,
    "wave_inleft":act.forwardleft,
    "wave_inclockwise":act.rotateright,
    "wave_inanticlockwise":act.rotateleft,
    "wave_inpunch":act.punchleft,
    "wave_out":act.reset,
    "wave_outup":"null",
    "wave_outdown":"null",
    "wave_outright":act.turnright,
    "wave_outleft":act.turnleft,
    "wave_outclockwise":"null",
    "wave_outanticlockwise":"null",
    "wave_outpunch":act.punchright
};

// array storing current gesture control setting
var matching = {};

// a empty array, use to emit data to client side
var array = [];

// catch the gesture event
Myo.on("pose_off", function(poseName){
    var lastPoseName = poseName;
  // binary tree structure classifier
    if(poseName !== "double_tap" && poseName !== "fingers_spread"){
        if(onTestPunch){
            lastPoseName = poseName + ' punch';
            onTestPunch = false;
        }else{
            onTestPunch = false;
            if(onTestPitch === "up"){
                lastPoseName = poseName + ' up';
            }else if(onTestPitch === "down"){
                lastPoseName = poseName + ' down';
            }
            if(onTestRoll === "anticlockwise" && onTestPitch === "none"){
                lastPoseName = poseName + ' anti-clockwise';
            }else if(onTestRoll === "clockwise" && onTestPitch === "none"){
                lastPoseName = poseName + ' clockwise';
            }
            if(onTestYaw === "left" && onTestPitch === "none" && onTestRoll === "none"){
                lastPoseName = poseName + ' left';
            }else if(onTestYaw === "right" && onTestPitch === "none" && onTestRoll === "none"){
                lastPoseName = poseName + ' right';
            }
        }
        
        io.emit("logSession", lastPoseName);
        io.in("logging").emit('instantLog', lastPoseName);
        
      //output result
        var gesture;
        
        switch(lastPoseName){
            case "fist":
                gesture = matching.fist
                break;
            case "fist up":
                gesture = matching.fistup
                break;
            case "fist down":
                gesture = matching.fistdown
                break;
            case "fist right":
                gesture = matching.fistright
                break;
            case "fist left":
                gesture = matching.fistleft
                break;
            case "fist clockwise":
                gesture = matching.fistclockwise
                break;
            case "fist anti-clockwise":
                gesture = matching.fistanticlockwise
                break;
            case "fist punch":
                gesture = matching.fistpunch
                break;
            case "wave_in":
                gesture = matching.wave_in
                break;
            case "wave_in up":
                gesture = matching.wave_inup
                break;
            case "wave_in down":
                gesture = matching.wave_indown
                break;
            case "wave_in right":
                gesture = matching.wave_inright
                break;
            case "wave_in left":
                gesture = matching.wave_inleft
                break;
            case "wave_in clockwise":
                gesture = matching.wave_inclockwise
                break;
            case "wave_in anti-clockwise":
                gesture = matching.wave_inanticlockwise
                break;
            case "wave_in punch":
                gesture = matching.wave_inpunch
                break;
            case "wave_out":
                gesture = matching.wave_out
                break;
            case "wave_out up":
                gesture = matching.wave_outup
                break;
            case "wave_out down":
                gesture = matching.wave_outdown
                break;
            case "wave_out right":
                gesture = matching.wave_outright
                break;
            case "wave_out left":
                gesture = matching.wave_outleft
                break;
            case "wave_out clockwise":
                gesture = matching.wave_outclockwise
                break;
            case "wave_out anti-clockwise":
                gesture = matching.wave_outanticlockwise
                break;
            case "wave_out punch":
                gesture = matching.wave_outpunch
                break;
            default:
        }
                
    }else{
        
        io.emit("logSession", poseName);
        io.in("logging").emit('instantLog', poseName);
        
        switch(poseName){
            case "double_tap":
                gesture = matching.double_tap
                break;
            case "fingers_spread":
                gesture = matching.fingers_spread
                break;
            default:
        }
       
    }
    // send to Recording tab
    if(isCalibrating && gesture !== act.play) {
        array.push(gesture);
        io.emit('instantCalibrate', gesture);
    }
    
  //send to Log tab
    if(isLogging){
        if(gesture === act.play) {
            test(array, function() {
                console.log('played array');
                i = 0;
            })
        }else {
            if(isControlling) {
                player.play(gesture, function(err){});
            }else {
                //console.log("isControlling === false");
            }
        }
    }
});

// determine punch motion
Myo.on('imu', function(data) {  
    var time = (new Date()).getTime();
    if (punchTime < time - 1000 && data.accelerometer.x < -2.0) {
        punchTime = time;
        onTestPunch = true;
    }
});

// determine arm motion and arm position
Myo.on('orientation', function(data){
    if(orientationRedo){
	   getFirstOrientation();
    }
	var orientationAngles ={
		roll : getRoll(data),
		pitch : getPitch(data),
		yaw : getYaw(data)
	};
	//console.log(firstOrientationAngles);
    //console.log(orientationAngles);
    
    if(orientationAngles.roll-firstOrientationAngles.roll > 2){
        onTestRoll = "anticlockwise";
    }else if(orientationAngles.roll-firstOrientationAngles.roll < -2){
        onTestRoll = "clockwise";
    }else{
        onTestRoll = "none";
    }
    
    if(orientationAngles.pitch > 15){
        onTestPitch = "up";
    }else if(orientationAngles.pitch < 2){
        onTestPitch = "down";
    }else{
        onTestPitch = "none";
    }
    
    if(firstOrientationAngles.yaw > 4.5 && firstOrientationAngles.yaw < 13.5){
        var angleChanged = orientationAngles.yaw - firstOrientationAngles.yaw;
        if(angleChanged > 2){
            onTestYaw = 'left';
        }else if(angleChanged < -2){
            onTestYaw = 'right';
        }else{
            onTestYaw = 'none';
        }
    }else if(firstOrientationAngles.yaw < 4.5){
        var angleChanged = orientationAngles.yaw - firstOrientationAngles.yaw;
        if(angleChanged > 2 && angleChanged < 4.5){
            onTestYaw = 'left';
        }else if(angleChanged < -2 || angleChanged > 13.5){
            onTestYaw = 'right';
        }else{
            onTestYaw = 'none';
        }
    }else if(firstOrientationAngles.yaw > 13.5){
        var angleChanged = orientationAngles.yaw - firstOrientationAngles.yaw;
        if(angleChanged < -2 && angleChanged > -4.5){
            onTestYaw = 'right';
        }else if(angleChanged > 2 || angleChanged < -13.5){
            onTestYaw = 'left';
        }else{
            onTestYaw = 'none';
        }
    }
    
    console.log('roll: ' + onTestRoll);
    console.log('pitch: ' + onTestPitch);
    console.log('yaw: ' + onTestYaw);
    
});

// record the initial position of arm
function getFirstOrientation(){
	firstOrientationAngles ={
		roll : getRoll(Myo.myos[0].lastIMU.orientation),
		pitch : getPitch(Myo.myos[0].lastIMU.orientation),
		yaw : getYaw(Myo.myos[0].lastIMU.orientation)
	};
	for (var i = 0; i < 100; i++) {
		firstOrientationAngles.roll  += getRoll(Myo.myos[0].lastIMU.orientation);
		firstOrientationAngles.pitch += getPitch(Myo.myos[0].lastIMU.orientation);
		firstOrientationAngles.yaw   += getYaw(Myo.myos[0].lastIMU.orientation); 
	};

	firstOrientationAngles.roll /= 100;
	firstOrientationAngles.pitch /= 100;
	firstOrientationAngles.yaw /= 100;

    orientationRedo = false;
}

// calculate roll angle
function getRoll(data){
	var roll = Math.atan2(2.0 * (data.w * data.x + data.y * data.z), 1.0 - 2.0 * (data.x * data.x + data.y * data.y));
	var roll_w = ((roll + Math.PI)/(Math.PI * 2.0) * 18);
	return roll_w;
}

// calculate pitch angle
function getPitch(data){
	var pitch = Math.asin(Math.max(-1.0, Math.min(1.0, 2.0 * (data.w * data.y - data.z * data.x))));
	var pitch_w = ((pitch + Math.PI/2.0)/Math.PI * 18);
	return pitch_w;
}

// calculate yaw angle
function getYaw(data){
	var yaw = Math.atan2(2.0 * (data.w * data.z + data.x * data.y), 1.0 - 2.0 * (data.y * data.y + data.z * data.z));
	var yaw_w = ((yaw + Math.PI)/(Math.PI*2.0) * 18);
	return yaw_w;
}

var temp = {};
// play a sequence of sound file
var i = 0;
function test(a, done){
    if(i < a.length){
        player.play(a[i], function(err){
            i++;
            test(a, done);
        });
    }else{
        done();
    }
}

io.on('connection', function(socket){
  console.log('a user connected');
// handle status request
  socket.on('statusReq', function(){
  	console.log('user asked for status');
  	temp.info = "Response for status";
  	var data = Myo.myos[0];
  	//console.log(data);
    Myo.trigger("zeroOrientation");
      
    isCalibrating = false;
    isLogging = false;
    
    var str = '';
    for (var p in matching) {
        if (matching.hasOwnProperty(p)) {
            str += p + ':' + matching[p] + '<br>';
            str = str.replace("./sound/", "");
            str = str.replace(".wav", "");
        }
    }

  	socket.emit('status', temp, data);
    socket.emit('showmenu', str);
  });
// handle record request
  socket.on('calibrationReq', function(){
  	console.log('user asked for calibration');
  	temp.info = "Response for calibration"; 
    
    isCalibrating = true;
    isLogging = false;
 
  	socket.emit('calibration', temp, array);
  });
// handle log request
  socket.on('logReq', function(roomData){
  	console.log('user asked for log');
  	temp.info = "Response for log";
      
    isCalibrating = false;
    isLogging = true;
    
    var str = '';
    for (var p in matching) {
        if (matching.hasOwnProperty(p)) {
            str += p + ':' + matching[p] + '<br>';
            str = str.replace("./sound/", "");
            str = str.replace(".wav", "");
        }
    }
    socket.emit('showmenu', str);
  	socket.emit('log', temp);
  	socket.join(roomData.room);
  });
// handle customization request
  socket.on('matchingReq', function(){
  	console.log('user asked for matching');
  	temp.info = "Response for matching";
      
    isCalibrating = false;
    isLogging = false;

  	socket.emit('matching', temp);
  });
// handle description request
  socket.on('descriptionReq', function(){
    console.log('user asked for description');
    isCalibrating = false;
    isLogging = false;
    socket.emit('description');
  });
  
// handle terminate
  socket.on('leaveLogging', function(){
    isControlling = false;
  	socket.leave("logging");
  });

  socket.on('disconnect', function(){
    console.log('user disconnected');
  });
// handle clean request
  socket.on('storeAction', function(actionData) {
      console.log('received data: ' + actionData.action);
      if(actionData.action === "clean") {
          array = [];
          console.log('storeAction: cleared array');
      }else {
          array.push(act[actionData.action]);
          console.log(array);
      }
  });
// handle start and stop request
  socket.on('operation', function(actionData) {
      if(actionData.action === "start") {
          isControlling = true;
      }else {
          isControlling = false;
      }
  });
// handle confirm and default request
  socket.on('matching', function(actionData, matchingArray){
      if(actionData === "send") {
          for(var k in matchingArray) {
            if(matchingArray.hasOwnProperty(k)) {
              switch(matchingArray[k]){
                case "forward":
                  matching[k] = act.forward;
                  break;
                case "forwardleft":
                  matching[k] = act.forwardleft;
                  break;
                case "forwardright":
                  matching[k] = act.forwardright;
                  break;
                case "backward":
                  matching[k] = act.backward;
                  break;
                case "backwardleft":
                  matching[k] = act.backwardleft;
                  break;
                case "backwardright":
                  matching[k] = act.backwardright;
                  break;
                case "bowup":
                  matching[k] = act.bowup;
                  break;
                case "bowdown":
                  matching[k] = act.bowdown;
                  break;
                case "punchleft":
                  matching[k] = act.punchleft;
                  break;
                case "punchright":
                  matching[k] = act.punchright;
                  break;
                case "open":
                  matching[k] = act.open;
                  break;
                case "close":
                  matching[k] = act.close;
                  break;
                case "rotateleft":
                  matching[k] = act.rotateleft;
                  break;
                case "rotateright":
                  matching[k] = act.rotateright;
                  break;
                case "rotatestop":
                  matching[k] = act.rotatestop;
                  break;
                case "turnleft":
                  matching[k] = act.turnleft;
                  break;
                case "turnright":
                  matching[k] = act.turnright;
                  break;
                case "reset":
                  matching[k] = act.reset;
                  break;
                case "play":
                  matching[k] = act.play;
                  break;
                default:
                  matching[k] = "null";
                  break;
              }
              
            }
          }
              
          console.log("matching: send");
      }else {
          console.log("matching: default");
          matching = Object.assign({}, matchingDefault);
          for(var k in matching){
            if(matching.hasOwnProperty(k)) {
              matching[k] = matchingDefault[k];
            }
          }
      }
  });
});

app.get('/', function(req, res){
	res.sendFile(__dirname + '/views/index.html');
});
